/*****************************************************************************/
/*
 * Linux driver for Easylon USB VNI Interfaces
 *
 * Copyright (c)2003 Gesytec GmbH, Volker Schober (info@gesytec.de)
 *
 */
 /*****************************************************************************/

#include "lonusb.h"

#include "dynbuf.c"

#include "driverdata.c"
#include "proc.c"
#include "ezusb.c"
#include "readwrite.c"
#include "fpga.c"
#include "driverentry.c"

#include "XO2_api.cpp"
#include "XO2_cmds.cpp"
#include "XO2_dev.cpp"

/*-------------------------------------------------------------------*/
/**
 *      DumpBuffer
 */
static void DumpBuffer(int DebugZone, const char *Prefix, int size,
		       const unsigned char *data)
{
	if (debug & DebugZone)
		print_hex_dump(KERN_DEBUG, Prefix, DUMP_PREFIX_OFFSET, 32, 1,
			       data, size, false);
}

static void dump_urb(struct urb *purb)
{
	dbg("urb                   :%p", purb);
	dbg("dev                   :%p", purb->dev);
	dbg("pipe                  :%08X", purb->pipe);
	dbg("status                :%d", purb->status);
	dbg("transfer_flags        :%08X", purb->transfer_flags);
	dbg("transfer_buffer       :%p", purb->transfer_buffer);
	dbg("transfer_buffer_length:%d", purb->transfer_buffer_length);
	dbg("actual_length         :%d", purb->actual_length);
	dbg("setup_packet          :%p", purb->setup_packet);
	dbg("start_frame           :%d", purb->start_frame);
	dbg("number_of_packets     :%d", purb->number_of_packets);
	dbg("interval              :%d", purb->interval);
	dbg("error_count           :%d", purb->error_count);
	dbg("context               :%p", purb->context);
	dbg("complete              :%p", purb->complete);
}

/*-------------------------------------------------------------------*/

/**
 *      lonusb_delete
 */
static void lonusb_delete(PDEVICE_EXTENSION pdx)
{
	int i;
	int dev_offset = pdx->minor * MAX_DEVICENODES;

	//printk(KERN_DEBUG "lonusb_delete(dev=%p): pSubDev=%p SubDevices=%p\n", dev, dev->pSubDev, dev->SubDevices);

	//if (dev->pSubDev)
	{
		for (i = 0; i < MAX_DEVICENODES; i++) {
			qexit(&pdx->pSubDev[i].ReadBufferQueue);

			device_destroy(lonusb_dev_class,
				       first_device + dev_offset + i);
		}

		//kfree(dev->pSubDev);
	}

//    if (dev->SubDevices)
//        kfree(dev->SubDevices);
	kfree(pdx);

	minor_table[pdx->minor] = NULL;
	if (pdx->intr_in_buff != NULL)
		kfree(pdx->intr_in_buff);
	if (pdx->bulk_out_buffer != NULL)
		kfree(pdx->bulk_out_buffer);
	if (pdx->write_urb != NULL) {
		dbg("usb_free_urb(dev->write_urb)");
		usb_free_urb(pdx->write_urb);
	}
	if (pdx->read_urb != NULL) {
		dbg("usb_free_urb(dev->read_urb)");
		usb_free_urb(pdx->read_urb);
	}

	pdx->RemainDataSize = 0;
}

/**
 *      lonusb_open
 */
static int lonusb_open(struct inode *inode, struct file *file)
{
	int subminor;
	int subdev;
	int submajor;
	int retval = 0;
	PDEVICE_EXTENSION pdx = NULL;

	dbgz(DEBUG_OPEN_CLOSE, "%s(): MINOR=%d", __FUNCTION__,
	     MINOR(inode->i_rdev));

	// fix minor_table index
	submajor = imajor(inode);
/*
#ifdef CONFIG_USB_DYNAMIC_MINORS
    // see minor_base at http://www.makelinux.net/ldd3/chp-13-sect-4  
    subminor = iminor(inode);
#else
    subminor = iminor(inode) - lonusb_class.minor_base;
#endif
*/
	if (iminor(inode) >= lonusb_class.minor_base) {
		subminor = iminor(inode) - lonusb_class.minor_base;
	} else {
		subminor = iminor(inode);
	}
	dbg("subminor = %i", subminor);

	if (submajor == MAJOR(first_device)) {
		subdev = subminor / MAX_DEVICENODES;

		dbg("attempt to open device: (%d, lvu_%d) %d:%d", submajor,
		    subminor, subdev, subminor % MAX_DEVICENODES);

		subminor = subminor % MAX_DEVICENODES;
	} else {
		err("usb device entry disabled. use standard character devices. current major number is %d", MAJOR(first_device));

		return -ENODEV;
	}

	if ((subdev < 0) || (subdev >= MAX_DEVICES) ||
	    (subminor < 0) || (subminor >= MAX_DEVICENODES))
		return -ENODEV;

	dbg("minor_table lookup");
	/* lock our minor table and get our local data for this minor */
	mutex_lock(&minor_table_mutex);
	pdx = minor_table[subdev];
	if (pdx == NULL) {
		mutex_unlock(&minor_table_mutex);
		return -ENODEV;
	}

	if (0 < pdx->pSubDev[subminor].open_count) {
		mutex_unlock(&minor_table_mutex);

		return -EBUSY;
	}

	dbg("minor_table lookup ok");

	mutex_unlock(&minor_table_mutex);

	/* lock this device */
	mutex_lock(&pdx->mutex);
	/* unlock the minor table */
	//dbg("usage count test ok");

	if (retval) {
		return -ERESTARTSYS;
	}

	/* increment our usage count for the driver */
	++(pdx->pSubDev[subminor].open_count);

	++pdx->sub_open_count;

	/* save our object in the file's private structure */
	file->private_data = &(pdx->pSubDev[subminor]);
	pdx->pSubDev[subminor].DevStats.open++;

	pdx->RemainDataSize = 0;
	qflush(&pdx->pSubDev[subminor].ReadBufferQueue);

	if (1 == pdx->sub_open_count) {
#if 0
		pdx->read_urb->status = 0;	// needed!
		pdx->read_urb->pdx = pdx->usbdev;	// needed!
#else
		pdx->read_urb->status = 0;	// needed!
		{
			int pipe =
			    usb_rcvintpipe(pdx->usbdev, pdx->intr_in_ep_addr);
			usb_fill_int_urb(pdx->read_urb, pdx->usbdev, pipe,
					 pdx->intr_in_buff, pdx->bulk_in_size,
					 lonusb_read_int_callback, pdx,
					 pdx->bInterval);
		}
#endif
		pdx->write_urb->dev = pdx->usbdev;	// needed!

		reset_neuron(pdx, FALSE, TRUE, POWER_UP_RESET);

		dbg("usb_submit_urb(read_urb=%p intr_in_ep_addr=%02X):",
		    pdx->read_urb, pdx->intr_in_ep_addr);
		dump_urb(pdx->read_urb);
		pdx->DevStats.submit_read1++;
		retval = USB_SUBMIT_URB(pdx->read_urb, GFP_KERNEL);
		if (retval) {
			pdx->DevStats.submit_read1_error++;
			err("%s(): usb_submit_urb(read_urb=%02X)=%d",
			    __FUNCTION__, pdx->intr_in_ep_addr, retval);
		} else
			pdx->DevStats.submit_read1_ok++;

		dump_urb(pdx->read_urb);

		if (retval < 0) {
			/* decrement our usage count for the device */
			--pdx->sub_open_count;
			--(pdx->pSubDev[subminor].open_count);
		}
	}

	/* unlock this device */
	mutex_unlock(&pdx->mutex);

	dbg("%s(): retval=%d open_count=%d", __FUNCTION__, retval,
	    pdx->sub_open_count);
	return retval;
}

/**
 *      lonusb_release
 */
static int lonusb_release(struct inode *inode, struct file *file)
{
	PSUB_DEVICE_EXTENSION dev = (PSUB_DEVICE_EXTENSION) file->private_data;
	int retval = 0;

	if (dev == NULL) {
		err("%s() object is NULL", __FUNCTION__);
		return -ENODEV;
	}

	dbgz(DEBUG_OPEN_CLOSE, "%s(): minor=%d open_count=%d (%d)",
	     __FUNCTION__,
	     (dev->udev->minor * MAX_DEVICENODES) + dev->SubdevNum,
	     dev->open_count, dev->udev->sub_open_count);

	/* lock our minor table */
	mutex_lock(&minor_table_mutex);

	/* lock our device */
	mutex_lock(&dev->udev->mutex);

	if (dev->open_count <= 0) {
		err("%s() device not opened", __FUNCTION__);
		retval = -ENODEV;
		goto exit_not_opened;
	}

	if ((dev->udev->usbdev == NULL) && (1 >= dev->udev->sub_open_count)) {
		/* the device was unplugged before the file was released */
		mutex_unlock(&dev->udev->mutex);
		lonusb_delete(dev->udev);
		mutex_unlock(&minor_table_mutex);
		return 0;
	}

	--(dev->udev->sub_open_count);

	--dev->open_count;

	/* decrement our usage count for the device */
	if (dev->udev->sub_open_count <= 0) {
		/* shutdown any bulk writes that might be going on */
		dump_urb(dev->udev->read_urb);
		usb_kill_urb(dev->udev->read_urb);
		dump_urb(dev->udev->read_urb);

		usb_kill_urb(dev->udev->write_urb);

		dev->udev->RemainDataSize = 0;
		qflush(&dev->ReadBufferQueue);
		dev->open_count = 0;

		// if (usb_reset_configuration(dev->usbdev) < 0)
		//     err("usb_reset_configuration() failed");
	}

exit_not_opened:
	mutex_unlock(&dev->udev->mutex);
	mutex_unlock(&minor_table_mutex);
	dev->DevStats.close++;

	return retval;
}
